import os
import time
import signal
import threading
from multiprocessing import Process
from multiprocessing import Manager
import unittest
from .base import *

def example_process():
    global stopflag
    stopflag = False

    def log(msg):
        with open("zdas.log", "a", encoding="utf-8") as fobj:
            print(msg, file=fobj)

    def on_exit(sig, frame):
        log(time.time())
        log("got term signal...")
        log(sig)
        log(frame)
        global stopflag
        stopflag = True
    signal.signal(signal.SIGTERM, on_exit)
    signal.signal(signal.SIGINT, on_exit)

    log("example_process start pid={}.".format(os.getpid()))
    while not stopflag:
        log(time.time())
        time.sleep(1)
    log("example_process end.")

def main03():
    daemon_start(example_process, "test03.pid", False)

def main04():
    daemon_start(example_process, "test04.pid", True)

class TestZdas(unittest.TestCase):

    def test01(self):
        p = Process(target=example_process)
        p.start()

        assert is_running(p.pid)

        p2 = get_process(p.pid)
        assert p2.pid == p.pid

        process_kill(p.pid)
        time.sleep(2)
        assert not p.is_alive()
        assert not is_running(p.pid)

    def test02(self):
        pidfile = "test02.pid"

        write_pidfile(pidfile)
        assert load_pid(pidfile) == os.getpid()

        clean_pid_file(pidfile)
        assert load_pid(pidfile) == 0

    def test03(self):
        pidfile = "test03.pid"

        p = Process(target=main03, daemon=True)
        p.start()

        time.sleep(1) # child process may not start yet
        pid = load_pid(pidfile)
        assert pid == p.pid
        assert is_running(pid)
        assert pid != os.getpid()

        print("killing pid={}".format(pid))
        daemon_stop(pidfile)
        time.sleep(2)
        p = get_process(pid)
        if p:
            print(p)
            print(p.status())
        assert not is_running(pid)

    def test04(self):
        pidfile = "test04.pid"

        p = Process(target=main04, daemon=True)
        p.start()

        time.sleep(1) # child process may not start yet
        pid = load_pid(pidfile)
        assert pid != p.pid
        assert pid != os.getpid()
        assert is_running(pid)

        print("killing pid={}".format(pid))
        daemon_stop(pidfile)
        time.sleep(2)
        p = get_process(pid)
        if p:
            print(p)
            print(p.status())
        assert not is_running(pid)



"""
Zencore daemon application start(zdas).
"""
import os
import errno
import atexit
import signal
import logging
import psutil

__all__  = [
    "process_kill",
    "load_pid",
    "write_pidfile",
    "get_process",
    "is_running",
    "clean_pid_file",
    "daemon_start",
    "daemon_stop",
]


logger = logging.getLogger(__name__)


def make_basic_daemon(workspace=None):
    """Make basic daemon.
    """
    workspace = workspace or os.getcwd()
    # first fork
    if os.fork():
        os._exit(0)
    # change env
    os.chdir(workspace)
    os.setsid()
    os.umask(0o22)
    # second fork
    if os.fork():
        os._exit(0)
    # reset stdin/stdout/stderr to /dev/null
    null = os.open('/dev/null', os.O_RDWR)
    try:
        for i in range(0, 3):
            try:
                os.dup2(null, i)
            except OSError as error:
                if error.errno != errno.EBADF:
                    raise
    finally:
        os.close(null)


def process_kill(pid, sig=None):
    """Send signal to process.
    """
    sig = sig or signal.SIGTERM
    os.kill(pid, sig)


def load_pid(pidfile):
    """read pid from pidfile.
    """
    if pidfile and os.path.isfile(pidfile):
        with open(pidfile, "r", encoding="utf-8") as fobj:
            return int(fobj.readline().strip())
    return 0


def write_pidfile(pidfile):
    """write current pid to pidfile.
    """
    pid = os.getpid()
    if pidfile:
        with open(pidfile, "w", encoding="utf-8") as fobj:
            fobj.write(str(pid))
    return pid


def get_process(pid):
    """get process information from pid.
    """
    try:
        return psutil.Process(pid)
    except psutil.NoSuchProcess:
        return None


def is_running(pid):
    """check if the process with given pid still running
    """
    process = get_process(pid)
    if process and process.is_running() and process.status() != "zombie":
        return True
    else:
        return False


def clean_pid_file(pidfile):
    """clean pid file.
    """
    if pidfile and os.path.exists(pidfile):
        os.unlink(pidfile)


def daemon_start(main, pidfile, daemon=True, workspace=None):
    """Start application in background mode if required and available. If not then in front mode.
    """
    logger.debug("start daemon application pidfile={} daemon={} workspace={}.".format(pidfile, daemon, workspace))
    new_pid = os.getpid()
    workspace = workspace or os.getcwd()
    os.chdir(workspace)
    daemon_flag = False
    if pidfile and daemon:
        old_pid = load_pid(pidfile)
        if old_pid:
            logger.debug("pidfile {} already exists, pid={}.".format(pidfile, old_pid))
        # if old service is running, just exit.
        if old_pid and is_running(old_pid):
            error_message = "Service is running in process: {}.".format(old_pid)
            logger.error(error_message)
            print(error_message, file=os.sys.stderr)
            os.sys.exit(95)
        # clean old pid file.
        clean_pid_file(pidfile)
        # start as background mode if required and available.
        if daemon and os.name == "posix":
            make_basic_daemon()
            daemon_flag = True
    if daemon_flag:
        logger.info("Start application in DAEMON mode, pidfile={} pid={}".format(pidfile, new_pid))
    else:
        logger.info("Start application in FRONT mode, pid={}.".format(new_pid))
    write_pidfile(pidfile)
    atexit.register(clean_pid_file, pidfile)
    main()
    return


def daemon_stop(pidfile, sig=None):
    """Stop application.
    """
    logger.debug("stop daemon application pidfile={}.".format(pidfile))
    pid = load_pid(pidfile)
    logger.debug("load pid={}".format(pid))
    if not pid:
        print("Application is not running or crashed...", file=os.sys.stderr)
        os.sys.exit(195)
    process_kill(pid, sig)
    return pid



import os
from setuptools import setup
from setuptools import find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.rst'), "r", encoding="utf-8") as fobj:
    long_description = fobj.read()

requires = [
    "psutil",
]

setup(
    name="zdas",
    version="0.1.0",
    description="Zencore daemon application start.",
    long_description=long_description,
    url="https://github.com/appstore-zencore/zdas.git",
    author="zencore",
    author_email="dobetter@zencore.cn",
    license="MIT",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords=['zdas'],
    packages=find_packages("src"),
    package_dir={"": "src"},
    zip_safe=False,
    requires=requires,
    install_requires=requires,
)